#!/usr/bin/perl

use strict;
use lib 'ck_';
use j_config;
use jury_common;

use util::logger;
use util::ipgiver;
use util::mailer;

use POSIX ':sys_wait_h';
use IO::Handle;
use Socket;

my $log;
my $log_bank;
my $log_brief;
my $log_ip;

my %childs;					# List of running PIDs
my %timeouts;					# List of PIDs, to whose SIGKILL was sent 
my %bad;

my $die_in_eval = 0;

++$|;		# No buffered output!

local $SIG{__DIE__} = sub
{
	if ( $die_in_eval == 0 )
	{
		# $^S - Current state of the interpreter.
		return unless defined $^S;		# State: 'Parsing module/eval'
		return if $^S;				# State: 'Executing an eval'
	}

	chomp( my $msg = shift );

	$log -> alert( "  *** Oops! PROCESS {$$} DIED" );
	$log -> alert( "  *** $msg" );
	for ( 0 .. 90 ) {
		my ( $pkg, $fname, $line, $sub ) = caller( $_ ) or last;
		$fname =~ s|^.*\\||;
		$fname =~ s|^.*/||;
		$log -> alert( sprintf "\t(%2d)  %-14s  %-25s  %s( )", $_, $pkg, "$fname:$line", $sub );
	}
	exit 1;
};

END {
	$log -> info( "   *** PROCESS FINISHED, $@, $!." );
	$log -> info( "" );
}

#
# ----------------------------- M A I N    C O D E    S T A R T -------------------------------
#

# DB connection

dbConnect( 'main' );
my @tids 	= getActiveTeams( );
my $gametime 	= getGameTime( );

# Create/open logfiles

$log = new util::logger( $LOG_MAIN )		or die "error: can't create '$LOG_MAIN': $!\n";
$log -> info ( "" );
print $log -> info ( "   ************************************ PROCESS STARTED" );

$log_ip = new util::logger( $LOG_IP )		or die "error: can't create '$LOG_IP': $!\n";
$log_ip -> info( '-'x80 );
$log_ip -> info( " start: getGameTime( ) = $gametime" );
$log_ip -> info( "config: IPGIVER_ENABLED: $IPGIVER_ENABLED" );

print $log -> info( "startup: teams: @tids" );
print $log -> info( "startup: getGameTime( ) = $gametime" );
print $log -> info( " config: child timeout = $TIMEOUT_CHILD sec" );

# ----------------------- create team logs 

my %tl;					# team logs
my %tbl;				# team brief logs

for ( @tids )
{
	$tl{$_} = new util::logger( sprintf $LOG_TEAMS_TEMPLATE, $_  )	or die "can't create team's '$_' log\n";
	$tl{$_} -> info( '' );
	$tl{$_} -> info( '*' x 110 );
	$tl{$_} -> info( "*** PROCESS STARTED: $gametime" );

	$tbl{$_} = new util::logger( sprintf $LOG_TEAMS_BRIEF, $_ )	or die "can't create team's '$_' brief log\n";
	$tbl{$_} -> info( '' );

	$tbl{$_} -> info( sprintf '----------- start: %-9s ----------------', $gametime );
}

# ---------------------- team logs created

doUnsetAlias();

warn " ---- sending newround \n";
my $saved;			# for temporary saving of main log
for my $tid ( @tids )
{
	( $saved, $log ) = ( $log, $tl{$tid} );		# fake
	$log_brief = $tbl{$tid};			# fake

	for my $servid ( @SERV_NEWROUND )
	{
		invoke( $TIMEOUT_NEW_ROUND, $servid, 'newRound', $tid, getGameTime( ) );
	}
	$log = $saved;					# restore
}
warn " ---- setting alias \n";
doSetRandAlias() if $IPGIVER_ENABLED;

for ( @tids ) 
{
	my $pid = fork( );

	$pid < 0 and 				die "cannot fork()!\n";

	if ( $pid )		# Parent process
	{
		$childs{$pid} = $_;
		print $log -> info( "   fork: pid '$pid', team '$_'" );
	}
	elsif ( $pid == 0 )	# Child process
	{
		$log_bank = new util::logger( sprintf $LOG_TEAMS_BANK, $_ ) or die "can't create team's '$_' bank log\n";
		$log_bank -> info( " g time: $gametime" );

		( $log, $log_brief ) = ( $tl{$_}, $tbl{$_} );

		dbConnect( $_ );
		doTeamCycle( $_ );

		exit 0;		# terminate child process
	}
}

# --- Parent process only !

$SIG{ALRM} = \&kill_all;
alarm $TIMEOUT_CHILD;

print $log -> info( "   wait: waiting for children to terminate ..." );
while ( 1 )
{
	my $kid = waitpid( -1, WNOHANG );

	if ( $kid > 0 )
	{
		print $log -> info( "   wait:    pid '$kid', team '$childs{$kid}' terminated $timeouts{$kid}" );
		delete $childs{$kid};
	}

	select( undef, undef, undef, 0.2 );
	last if $kid < 0;
}

print $log -> info( " finish: ALL children terminated!" );

# ---------------------------------------- increment game time

dbConnect( 'main' );

my $new_time = getGameTime( ) + $GAMETIME_ROUND;
print $log -> info( " finish: setGameTime( $new_time )" );
setGameTime( $new_time );

my $n = cleanReservedOrbits( );
print $log -> info( "  clean: $n old reserved orbits deleted" );

$n = cleanOldSputniks( );
print $log -> info( "  clean: $n old sputniks deleted" );

$n = cleanOldSauths( );
print $log -> info( "  clean: $n old sauths deleted" );

doUnsetAlias( );

#
# ------------------------------ S u b r o u t i n e s --------------------------------------
#

sub doTeamCycle
{
	my ( $tid ) = @_;

	#
	#	1. check( )	--------------------------------------------------
	#
	
	$log -> info( '----------------------------------- check ------------------------------------' );

	for ( @SERV_ALL )
	{
		/control/ and next;
		my ( $code, $comm ) = invoke( $TIMEOUT_CHECK, $_, 'check', $tid );
		
		$bad{$_} = "$code: general failure: $comm" if $code != 1;
	}

	#
	#	2. store( ) - standard services		--------------------------
	#

	$log -> info( '----------------------------------- store ------------------------------------' );

	for ( @SERV_STORE_STD )
	{
		my $fid = $tid . randStr( 19, 'aeocusiy' );
		store_flag( $tid, $_, $fid );
	}

	#
	# 	3. store( ) to 'control' and 'data'	--------------------------
	#
	
	my @sids = getSputniksToStore( );
	$log -> info( "     db: ". (+@sids) . " sputniks to store flags to" );
	for ( @sids )
	{
		store_flag( $tid, 'control', $_ );
		store_flag( $tid, 'data', $_ );
	}

	$log -> info( '----------------------------------- launch -----------------------------------' );

	#
	#	4. launch( ) normal satellites	----------------------------------
	#
	
	for ( 1 .. $SAT_PER_ROUND )
	{
		my ( $code, $comm ) = invoke( $TIMEOUT_LAUNCH, 'start', 'launch', $tid );
		$bad{start} = $bad{start} || "$code: launch failed: $comm" and next if $code != 1;

		my @ss = split ';', $comm;
		if ( @ss != 2 )
		{
			$log->warning( "   warn: \t\t\t\t\tBAD ANSWER FORMAT!" );
			next;
		}
		$die_in_eval = 0;
		eval {
			setSputnikAuth( $ss[0], $ss[1] );	# sid, sauth
		};
		$die_in_eval = 1;
	}

	#
	#	5. launch( ) alien satellites	----------------------------------
	#
	
	for ( 1 .. $MIL_PER_ROUND )
	{
		my ( $code, $comm ) = invoke( $TIMEOUT_LAUNCH, 'alien', 'launch', $tid );
		$bad{alien} = $bad{alien} || "$code: launch failed: $comm" and next if $code != 1;

		my @ss = split ';', $comm;					# sid, sauth, secret
		if ( @ss != 3 )
		{
			$log->warning( "   warn: \t\t\t\t\tBAD ANSWER FORMAT!" );
			next;
		}
		$die_in_eval = 0;
		eval {
			setSputnikAuth( $ss[0], $ss[1] );				# sid, sauth
		};
		$die_in_eval = 1;
		next if $@;

		my $flag = $ss[0] . ';' . $ss[2];
		addNewFlag( $tid, $ss[0], 'alien', $flag );
		$log -> info( "     db: + new flag: $flag" );
	}

	#
	#	6. check( ) control	------------------------------------------
	#

	my ( $code, $comm ) = invoke( $TIMEOUT_CHECK, 'control', 'check', $tid );
	$bad{control} = $bad{control} || "$code: post-launch check fail: $comm" if $code != 1;

	#
	#	7. retrieve( ) random flags	----------------------------------
	#
	
	$log -> info( '------------------------------ random retrieve -------------------------------' );
	
	for my $servid ( @SERV_RAND_RETR )
	{
		my @fids = getActiveFlags( $tid, $servid, getGameTime( ) - $LIFE_FLAG );
		$log -> info( sprintf "     db: %7s: %3s active flags", $servid, 0+@fids );

		my $fid_count = @fids	or next;

		for ( 1 .. $NUM_RAND_FLAGS )
		{
			my $i = int( rand $fid_count );
			my ( $fid, $flag ) = @{$fids[ $i ]};
			$log -> info( "   rand: # $i of $fid_count" );
		
			my ( $code, $str ) = invoke( $TIMEOUT_RETR, $servid, 'retrieve', $tid, $fid, $flag );
			$bad{$servid} = $bad{$servid} || "$code: rand flag check fail: $str" and last if $code != 1;

			# Team earned some money :)
		}
	}

	#
	#	8. Print out summary, add status to DB, send checks to teams
	#

	$log -> info( '---------------------------------- SUMMARY -----------------------------------' );
	$log_brief -> info( '------------------ summary -----------------' );
	for ( @SERV_ALL )
	{
		$bad{$_} = $bad{$_} || '1: OK';
		$bad{$_} =~ /^(\d+):\s*(.*)/;
		my ( $code, $msg ) = ( $1||0, $2||'parsing error: '.$bad{$_} );

		setServiceStatus( $tid, $_, $code, $msg );			# ( tid, servid, status, scomment )

		my $money = $STATUS_COST[ $code ];

		$log -> info( sprintf "summary: %-9s : %2d pts - '%s'", $_, $money, $bad{$_} );
		$log_brief -> info( ellipsis( sprintf " %-9s %2d pts - %s", $_, $money, ellipsis( $bad{$_} ) ) );

		if ( $money != 0 )
		{
			my $check = getCheck( $money );
			if ( $check ) {
				my $m = sendTeamMail( $tid, "Earn $money pts for service $_", "$check" );
				if ( $m == 0 ) {
					$log -> info( "   mail: sent ok" );
				}
				else {
					$log -> info( "   mail: FAIL: $!" );
				}

			}
			else {
				$log -> warning( "  check: getCheck( $money ) FAILED" );
			}
		}
	}
}

sub store_flag($$$)
{
	my ( $tid, $servid, $fid ) = @_;
	my $flag = '='.randStr( 63, $FLAG_ALF );		# FLAG GENERATION !!

	my ( $code, $str );

	( $code, $flag ) = invoke( $TIMEOUT_STORE, $servid, 'store', $tid, $fid, $flag );	
	$bad{$servid} = $bad{$servid} || "$code: store flag failure1: $flag" and return if $code != 1;

	( $code, $str ) = invoke( $TIMEOUT_RETR, $servid, 'retrieve', $tid, $fid, $flag );
	$bad{$servid} = $bad{$servid} || "$code: store flag failure2: $str" and return if $code != 1;

	$log -> info( "     db: + new flag: $flag" );
	addNewFlag( $tid, $fid, $servid, $flag );
}

# --------------------------------------------------------------------------------------------------------
#				 I N T E R N A L     S U B R O U T I N E S
# --------------------------------------------------------------------------------------------------------

sub doUnsetAlias()
{
	my $s = ipUnsetAlias() ? 'ok' : 'fail';
	$log_ip -> info( "unset: $s" );
	print $log -> info( "ipgiver: unset $s" );
}

sub doSetRandAlias()
{
	my $ip = ipSetRandAlias() || 'fail';
	$log_ip -> info( "  set: $ip" );
	print $log -> info( "ipgiver: set $ip" );
}

sub kill_all()
{
	for ( keys %childs )
	{
		kill 9, $_;
		$timeouts{$_} = '[timeout]';
	}
}

#
#	invoke - вызов процедуры из чекера.
#

sub invoke
{
        my ( $timeout, $servid, $sub, @args ) = @_;
        my @r = ( 0, "$sub( ) unexpected error" );

        s/'/\\'/g for @args;				# for security!

        local $" = "', '";
	my $eval = "$sub( '@args' )";

	$log -> info( sprintf " invoke: %7s %-s", $servid, $eval );	# this goes to team's log

	require "ck_$servid.pm";

	my $_bak = $_;		# :)

        eval qq{

		\@r = ( 0, 'internal error' );

		local \$SIG{ALRM} = sub
		{
			\$die_in_eval = 0;
			\@r = ( 4, 'timeout: $timeout sec' );
			die "timeout: $timeout sec";
		};
		alarm $timeout;

		\$die_in_eval = 1;	# die on unexpected errors inside checker
                use ck_$servid;
                \@r = $eval;
		\$die_in_eval = 0;

		alarm 0;
        };

	$_ = $_bak;		# :)

	s/[\r\n]//g for @r;		# some checkers may be not clean :)

	$log -> info( " invoke: \t\t\t\t\t'@r'" );	# too many '\t'? :)

	my $s = $r[0] == 1 ? 'OK' : $r[1];

	$log_brief -> info( sprintf " %7s %-8s  %-3s  %s", $servid, $sub, $r[0], ellipsis( $s, 60 ) );

	@r;
}

sub ellipsis($$)
{	
	my ( $s, $n ) = @_;
	$s =~ s/^(.{$n}).*/$1 .../;
	$s;
}

sub randStr($$)
{
	my $alf = pop;
	join '', map { substr $alf,int(rand(length($alf))),1 } 1..pop;
}

sub getCheck($)
{
	my $sum = pop;
	$log_bank -> info( "    sub: getCheck( $sum ) ..." );

	socket( S, PF_INET, SOCK_STREAM, 0 )		or do {
								$log_bank -> warning( "\t\tsocket: $!" );
								die "socket: $!\n";
							};
	my $addr = inet_aton $BANK_HOST			or do {
								$log_bank -> warning( "\t\tinet_aton: $!" );
								die "inet_aton: $!\n";
							};
	connect( S, sockaddr_in( $BANK_PORT, $addr ) )	or do {
								$log_bank -> warning( "\t\tconnect: $!" );
								die "connect: $!\n";
							};

	my $req = "get_check $sum";
	$log_bank -> info( "request: [S<-C] '$req'" );
	send( S, $req."\r\n", 0 );

	recv( S, my $reply, 1024, 0 );
	$reply =~ s/[\r\n]+$//;
	$log_bank -> info( "  reply: [S->C] '$reply'" );

	close S;

	if ( $reply =~ /^\+(\+[A-Za-z0-9_\/\+\-]+)/ )
	{
		$log_bank -> info( "\t\tOK" );
		return $1;
	}
	else
	{
		$log_bank -> warning( "\t\tREMOTE FAILURE!" );
		return undef;
	}
}


